"""This module implements a HttpStream using the requests lib."""
from typing import IO, Iterable, Tuple

import cytoolz
import dateutil.parser
import requests
import requests_toolbelt

from iotoolz._abc import AbcStream, StreamInfo

_to_datetime = cytoolz.excepts(Exception, dateutil.parser.parse)


class HttpStream(AbcStream):
    """HttpStream is the stream interface to http/https resources using "requests"."""

    supported_schemas = {"http", "https"}

    def read_to_iterable_(
        self, uri: str, chunk_size: int, fileobj: IO[bytes], **kwargs
    ) -> Tuple[Iterable[bytes], StreamInfo]:
        """
        Implements a http GET request using "requests" with stream flag always set.

        Args:
            uri (str): http or https endpoint to retrieve the data.
            chunk_size (int): size for each bytes chunk (provided in the Stream constructor).
            fileobj (IO[bytes]): file-like object of the stream buffer.
            **kwargs: keyword arguments that will be passed to the "requests.get" method.

        Returns:
            Tuple[Iterable[bytes], StreamInfo]: tuple of the bytes iterable and StreamInfo.
        """
        resp = requests.get(
            uri, stream=True, **cytoolz.dissoc(kwargs, "stream", "use_post")
        )
        resp.raise_for_status()
        info = StreamInfo(
            content_type=resp.headers.get("Content-Type"),
            encoding=resp.encoding,
            etag=resp.headers.get("etag"),
        )
        return resp.iter_content(chunk_size=chunk_size), info

    def write_from_fileobj_(
        self, uri: str, fileobj: IO[bytes], size: int, **kwargs
    ) -> StreamInfo:
        """
        Implements a http PUT or POST request using the "requests" lib.

        This implements takes in an additional "use_post" flag to indicate whether the
        http method should be PUT (default) or POST.

        Args:
            uri (str): http/https endpoint to PUT/POST the data.
            fileobj (IO[bytes]): file-like object of the current data in buffer.
            size (int): size of the data in the file-like object.
            **kwargs: additional keyword arguments to pass to the "requests.put" or "requests.post" method.

        Returns:
            StreamInfo: [description]
        """
        use_post = kwargs.get("use_post")
        headers = kwargs.get("headers", {})
        if "content-type" not in headers:
            headers["content-type"] = self.content_type
        if "charset=" not in headers["content-type"] and self.encoding:
            if not headers["content-type"].endswith(";"):
                headers["content-type"] += "; "
            headers["content-type"] += f"charset={self.encoding}"
        requests_method = requests.post if use_post else requests.put
        resp = requests_method(
            uri,
            data=requests_toolbelt.StreamingIterator(size, fileobj),
            headers=headers,
            **cytoolz.dissoc(kwargs, "use_post", "data", "headers"),
        )
        resp.raise_for_status()
        return StreamInfo(content_type=self.content_type, encoding=self.encoding)

    def stats_(self) -> StreamInfo:
        resp = requests.head(
            self.uri, **cytoolz.dissoc(self._kwargs, "stream", "use_post")
        )
        resp.raise_for_status()
        last_modified_str = resp.headers.get("Last-Modified")
        last_modified = _to_datetime(last_modified_str) if last_modified_str else None
        return StreamInfo(
            content_type=resp.headers.get("Content-Type"),
            encoding=resp.encoding,
            etag=resp.headers.get("ETag"),
            last_modified=last_modified,
        )

    def unlink(self, missing_ok: bool = True, **kwargs):
        kwargs = {**self._kwargs, **kwargs}
        try:
            resp = requests.delete(
                self.uri, **cytoolz.dissoc(kwargs, "stream", "use_post")
            )
            resp.raise_for_status()
        except Exception:  # pylint: disable=broad-except
            if not missing_ok:
                raise

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False,
    ):
        """This method does nothing as the actual HTTP call will handle any folder creation as part of the request."""
        ...

    def iter_dir_(self) -> Iterable[StreamInfo]:
        """This method does nothing."""
        return ()

    def exists(self) -> bool:
        """Whether the http resource exists."""
        resp = requests.head(
            self.uri, **cytoolz.dissoc(self._kwargs, "stream", "use_post")
        )
        return resp.ok



"""This module implements the FileStream with python native "open" method."""
import datetime
import io
import os
import os.path
import pathlib
import shutil
from typing import IO, Iterable, Tuple, Union

from iotoolz._abc import AbcStream, StreamInfo
from iotoolz.utils import guess_content_type_from_file


class FileStream(AbcStream):
    """FileStream is the stream interface to the local file system with python's "open" method."""

    supported_schemas = {"", "file"}

    def __init__(
        self,
        uri: Union[str, pathlib.Path],
        mode: str = "r",
        buffering: int = -1,
        encoding: str = None,
        newline: str = None,
        content_type: str = "",
        inmem_size: int = None,
        delimiter: Union[str, bytes] = None,
        chunk_size: int = io.DEFAULT_BUFFER_SIZE,
        etag: str = "",
        **kwargs,
    ):
        super().__init__(
            str(uri),
            mode,
            buffering,
            encoding,
            newline,
            content_type,
            inmem_size,
            delimiter,
            chunk_size,
            etag,
            **kwargs,
        )

    def read_to_iterable_(
        self, uri: str, chunk_size: int, fileobj: IO[bytes], **kwargs
    ) -> Tuple[Iterable[bytes], StreamInfo]:
        last_modified = datetime.datetime.fromtimestamp(
            pathlib.Path(self.uri).stat().st_mtime
        )

        def iter_bytes() -> Iterable[bytes]:
            with open(
                self.uri, mode="rb", buffering=self.buffering, newline=self.newline,
            ) as stream:
                for chunk in stream:
                    yield chunk

        return (
            iter_bytes(),
            StreamInfo(
                content_type=guess_content_type_from_file(self.uri),
                last_modified=last_modified,
            ),
        )

    def write_from_fileobj_(
        self, uri: str, fileobj: IO[bytes], size: int, **kwargs
    ) -> StreamInfo:
        os.makedirs(os.path.dirname(uri), exist_ok=True)
        mode = self.mode.replace("r", "")
        if "b" not in mode:
            mode += "b"
        with open(
            self.uri,
            mode=mode,
            buffering=self.buffering,
            encoding=None,
            newline=self.newline,
        ) as stream:
            shutil.copyfileobj(fileobj, stream)
        last_modified = datetime.datetime.fromtimestamp(
            pathlib.Path(self.uri).stat().st_mtime
        )
        return StreamInfo(last_modified=last_modified)

    def stats_(self) -> StreamInfo:
        last_modified = datetime.datetime.fromtimestamp(
            pathlib.Path(self.uri).stat().st_mtime
        )
        return StreamInfo(
            content_type=guess_content_type_from_file(self.uri),
            last_modified=last_modified,
        )

    def unlink(self, missing_ok: bool = True, **kwargs):
        try:
            pathlib.Path(self.uri).unlink()
        except FileNotFoundError:
            if not missing_ok:
                raise

    def is_dir(self) -> bool:
        """Whether stream points to a existing dir."""
        return pathlib.Path(self.uri).is_dir()

    def is_file(self) -> bool:
        """Whether stream points to a existing file."""
        return pathlib.Path(self.uri).is_file()

    def rmdir(self, ignore_errors: bool = False, **kwargs) -> "FileStream":
        """Remove the entire directory."""
        print(self.uri)
        shutil.rmtree(self.uri, ignore_errors=ignore_errors, **kwargs)
        return self

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False,
    ):
        """
        Create a new directory at this given path. If mode is given, it is combined with
        the process’ umask value to determine the file mode and access flags. If the path
        already exists, FileExistsError is raised.

        If parents is true, any missing parents of this path are created as needed; they
        are created with the default permissions without taking mode into account
        (mimicking the POSIX mkdir -p command).

        If parents is false (the default), a missing parent raises FileNotFoundError.

        If exist_ok is false (the default), FileExistsError is raised if the target
        directory already exists.

        If exist_ok is true, FileExistsError exceptions will be ignored (same behavior
        as the POSIX mkdir -p command), but only if the last path component is not an
        existing non-directory file.

        Args:
            mode (int, optional): mask mode. Defaults to 0o777.
            parents (bool, optional): If true, creates any parents if required. Defaults to False.
            exist_ok (bool, optional): If true, will not raise exception if dir already exists. Defaults to False.
        """
        pathlib.Path(self.uri).mkdir(mode=mode, parents=parents, exist_ok=exist_ok)

    def iter_dir_(self) -> Iterable[StreamInfo]:
        dirpath = self.uri if os.path.isdir(self.uri) else os.path.dirname(self.uri)
        for entry in os.scandir(dirpath):
            yield StreamInfo(
                uri=entry.path,
                name=entry.name,
                last_modified=datetime.datetime.fromtimestamp(entry.stat().st_mtime),
            )

    def exists(self) -> bool:
        """Whether the stream points to an existing resource."""
        return pathlib.Path(self.uri).exists()



"""
Common io utils  based on existing libs.
"""
import contextlib
import functools
import io
import os.path
from typing import IO, Any, Iterable, Iterator, Optional, Tuple, TypeVar

import cytoolz
import magic
from chardet.universaldetector import UniversalDetector

T = TypeVar("T", io.IOBase, IO, Any)


@contextlib.contextmanager
def peek_stream(
    stream: T, peek: Optional[int] = None, ignore_closed: bool = True
) -> Iterator[T]:
    """
    Context manager to restore the stream position when exiting the context.

    If the arg "peek" is provided, stream will start at the provided position when
    entering the context.

    Args:
        stream (T): Any stream object with the seek and tell method.
        peek (Optional[int], optional): stream position to start at when entering the context. Defaults to None.
        ignore_closed (bool, optional): do not restore to position if file is already closed. Defaults to True.

    Raises:
        TypeError: stream is not seekable.

    Returns:
        Iterator[T]: stream object with the modified position.
    """
    if not hasattr(stream, "seek") or not hasattr(stream, "tell"):
        raise TypeError(f"{stream} is not seekable")
    pos = stream.tell()  # type: ignore
    try:
        if isinstance(peek, int):
            stream.seek(peek)  # type: ignore
        yield stream
    finally:
        if not (ignore_closed and stream.closed):
            stream.seek(pos)  # type: ignore


def guess_encoding(
    data: Iterable[bytes], default_encoding: str = "utf-8"
) -> Tuple[str, float]:
    """
    Guess the encoding to decode bytes into corresponding string object.

    Uses chardet to attempt to progressively guess the encoding which can be used to
    decode the bytes into corresponding strings. Returns a tuple[encoding, confidence].

    Args:
        data (Iterable[bytes]): [description]
        default_encoding (str, optional): [description]. Defaults to "utf-8".

    Returns:
        Tuple[str, float]: [description]
    """

    detector = UniversalDetector()
    for line in data:
        detector.feed(line)
        if detector.done:
            break
    detector.close()
    return (
        detector.result.get("encoding", default_encoding),  # type: ignore
        detector.result.get("confidence", 0.0),
    )


def guess_filename(uri: str) -> str:
    if uri.endswith("/"):
        return ""
    return os.path.basename(uri)


guess_content_type_from_file = cytoolz.excepts(
    IOError, functools.partial(magic.from_file, mime=True), lambda _: ""
)


guess_content_type_from_buffer = cytoolz.excepts(
    Exception, functools.partial(magic.from_buffer, mime=True), lambda _: ""
)



"""
`iotoolz.streams` provide a `Streams` helper class for creating stream objects.
"""
import functools
import io
import pathlib
import urllib.parse
from typing import Any, Dict, Iterable, Iterator, List, Type, Union

from iotoolz._abc import AbcStream, StreamInfo
from iotoolz.extensions import S3Stream
from iotoolz.file import FileStream
from iotoolz.http import HttpStream
from iotoolz.temp import TempStream

AbcStreamType = Type[AbcStream]

DEFAULT_STREAMS: List[AbcStreamType] = [FileStream, TempStream, HttpStream, S3Stream]


class Streams:
    """
    Helper class to open streams with the corresponding AbcStream implementation
    based on the uri schema.
    """

    INMEM_SIZE: int = 0

    def __init__(self, *stream_types: AbcStreamType):
        """
        Creates a new instance of Streams.
        """
        self._schema2stream: Dict[str, AbcStreamType] = {
            "": FileStream,
            "file": FileStream,
        }
        self._schema2kwargs: Dict[str, dict] = {}
        for stream_type in stream_types:
            self.register(stream_type)

        # alias
        self.open_stream = self.open

    def set_schema_kwargs(self, schema: str, **kwargs) -> "Streams":
        """
        Set and update the default parameters for the different AbcStream implementations
        based on their schema.

        ```py
        # do not verify ssl cert for all https requests
        # by passing verify=False flag to requests inside HttpStream
        # i.e. HttpStream is implemented with requests
        Streams().set_schema_kwargs("https", verify=False).open("https://foo.bar")
        ```

        Returns:
            Streams: current Streams object.
        """
        self._schema2kwargs[schema] = kwargs
        return self

    def register(
        self,
        stream_type: AbcStreamType,
        schemas: Union[str, Iterable[str]] = None,
        **kwargs,
    ) -> "Streams":
        """
        Register an concrete AbcStream (e.g. HttpStream, FileStream, TempStream) and the
        corresponding schemas to the Streams object.

        Steams will use the appropriate registered concrete AbcStream based on the uri
        schema.

        By default, FileStream will already be registered as it is the fallback for
        any unknown uri.

        Raises:
            ValueError: schema is not provided and {stream_type} has an empty supported_schemas.

        Returns:
            Streams: current Streams object.
        """
        schemas = schemas or stream_type.supported_schemas
        if schemas is None:
            raise ValueError(
                f"schema is not provided and {stream_type} has an empty supported_schemas."
            )
        schemas = [schemas] if isinstance(schemas, str) else schemas
        for schema in schemas:
            self._schema2stream[schema] = stream_type
            self._schema2kwargs[schema] = kwargs or self._schema2kwargs.get(schema, {})
        return self

    def open(  # pylint: disable=too-many-arguments
        self,
        uri: Union[str, pathlib.Path],
        mode: str = "r",
        schema: str = None,
        data: Union[bytes, str] = None,
        fileobj: Any = None,
        buffering: int = -1,
        encoding: str = None,
        newline: str = None,
        content_type: str = "",
        inmem_size: int = None,
        delimiter: Union[str, bytes] = None,
        chunk_size: int = io.DEFAULT_BUFFER_SIZE,
        etag: str = "",
        schema_kwargs: dict = None,
        **extra_kwargs,
    ) -> AbcStream:
        """
        Open an appropriate registered stream based on the uri schema.

        For example, if HttpStream is registered for http/https schema, any uri starting
        with http or https will be opended using HttpStream.

        You can perform most IO methods on the stream (e.g. read, write, seek, ...)
        depending on the provided mode (e.g. r, rb, w, wb, ...).

        If you open the stream with a context, it will close automatically after exiting
        the context.

        The only difference with a native "open" is that changes to the stream buffer
        will not be pushed to the actual remote resource until "save" or "close" is
        called.

        Appropriate concrete AbcStreams must be registered with Streams to support the
        corresponding uris.

        ```py
        streams = Streams().register(HttpStream, {"http", "https"})

        # Get some ndjson data from a https endpoint
        with streams.open("https://foo.bar/data.ndjson", "r") as stream:
            ndjsons = [json.loads(line) for line in stream]

        # Post some binary content to a http endpoint
        with streams.open("https://foo.bar/api/data", "wb") as stream:
            stream.write(b"hello world")

        ```

        Args:
            uri (Union[str, pathlib.Path]): uri to the resource.
            mode (str, optional): same as "open". Defaults to "r".
            schema (str, optional): if provided, supersed the schema inferred from the uri. Defaults to None.
            data (Union[bytes, str], optional): if provided, create a stream with the initial data. Defaults to None.
            fileobj (Any, optional): if provided, create a stream with the data copied from the fileobj. Defaults to None.
            buffering (int, optional): same as "open". Defaults to -1.
            encoding (str, optional): encoding used to decode binary data to string. Defaults to None.
            newline (str, optional): same as "open". Defaults to None.
            content_type (str, optional): resource mime type (e.g. application/json). Defaults to "".
            inmem_size (int, optional): max data size before buffer is rollover into disk. Defaults to None (i.e. never - may result in MemoryError).
            delimiter (Union[str, bytes], optional): delimiter used for determining line boundaries. Defaults to None.
            chunk_size (int, optional): size of chunks to return in binary mode. Defaults to io.DEFAULT_BUFFER_SIZE.
            etag (str, optional): etag for the stream content. Defaults to "".
            schema_kwargs (dict, optional): dict of schema to default parameters for the corresponding AbcStreams. Defaults to None.

        Returns:
            AbcStream: concrete AbcStream based on the uri schema.
        """
        if isinstance(uri, pathlib.Path):
            schema = "file"

        if not schema:
            schema, *_ = urllib.parse.urlsplit(str(uri))

        # fallback to file system
        if schema not in self._schema2stream:
            schema = "file"

        schema_kwargs = schema_kwargs or {}
        kwargs = {
            **self._schema2kwargs.get(schema, {}),
            **schema_kwargs.get(schema, {}),
            **extra_kwargs,
        }
        stream = self._schema2stream[schema].open(
            str(uri),
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            newline=newline,
            content_type=content_type,
            inmem_size=inmem_size or self.INMEM_SIZE,
            delimiter=delimiter,
            chunk_size=chunk_size,
            etag=etag,
            **kwargs,
        )

        if data:
            # temp make it writable
            stream.mode = "wb"
            stream.write(data)
            # reset to start pos
            stream.seek(0)
            #  revert to original
            stream.mode = mode

        elif fileobj and hasattr(fileobj, "read"):
            read = functools.partial(fileobj.read, chunk_size)
            if "b" in mode:
                iter_data = iter(read, b"")
            else:
                iter_data = iter(read, "")
            # temp make it writable
            stream.mode = "wb"
            for chunk in iter_data:
                stream.write(chunk)
                # in case the mode is wrong and iter doesn't end
                if not chunk:
                    break
            # reset to start pos
            stream.seek(0)
            #  revert to original
            stream.mode = mode

        return stream

    def mkdir(
        self,
        uri: Union[pathlib.Path, str],
        mode: int = 0o777,
        parents: bool = False,
        exist_ok: bool = False,
    ):
        """
        Create a new directory at this given path. If mode is given, it is combined with
        the process’ umask value to determine the file mode and access flags. If the path
        already exists, FileExistsError is raised.

        If parents is true, any missing parents of this path are created as needed; they
        are created with the default permissions without taking mode into account
        (mimicking the POSIX mkdir -p command).

        If parents is false (the default), a missing parent raises FileNotFoundError.

        If exist_ok is false (the default), FileExistsError is raised if the target
        directory already exists.

        If exist_ok is true, FileExistsError exceptions will be ignored (same behavior
        as the POSIX mkdir -p command), but only if the last path component is not an
        existing non-directory file.

        Args:
            uri (Union[pathlib.Path, str]): uri to create dir.
            mode (int, optional): mask mode. Defaults to 0o777.
            parents (bool, optional): If true, creates any parents if required. Defaults to False.
            exist_ok (bool, optional): If true, will not raise exception if dir already exists. Defaults to False.
        """
        return self.open(uri).mkdir(mode, parents, exist_ok)

    def iter_dir(self, uri: Union[pathlib.Path, str]) -> Iterator[AbcStream]:
        """
        If a directory, yields stream in the directory. Otherwise, yield all streams in
        the same directory as the provided uri.

        Args:
            uri (Union[pathlib.Path, str]): uri to list.

        Returns:
            Iterator["AbcStream"]: iterator of streams that matches the pattern.

        Yields:
            AbcStream: stream object
        """
        return self.open(uri).iter_dir()

    def glob(self, uri: Union[pathlib.Path, str]) -> Iterator[AbcStream]:
        """
        Yield streams that matches the provided scheme and pattern.

        Args:
            uri (Union[pathlib.Path, str]): base uri to glob.
            pattern (str, optional): unix shell style pattern. Defaults to "*".

        Returns:
            Iterator["AbcStream"]: iterator of streams that matches the pattern.

        Yields:
            AbcStream: stream object
        """
        uri_ = str(uri)
        first_star = uri_.index("*")
        return self.open(uri_[0:first_star]).glob(uri_[first_star:])

    def exists(self, uri: Union[pathlib.Path, str, AbcStream]) -> bool:
        """Whether a stream points to an existing resource."""
        if isinstance(uri, AbcStream):
            return uri.exists()
        return self.open(uri).exists()

    def stats(self, uri: Union[pathlib.Path, str, AbcStream]) -> StreamInfo:
        """Get the StreamInfo."""
        if isinstance(uri, AbcStream):
            return uri.stats()
        return self.open(uri).stats()

    def unlink(
        self,
        uri: Union[pathlib.Path, str, AbcStream],
        missing_ok: bool = True,
        **kwargs,
    ):
        """Delete and remove a stream resource."""
        if isinstance(uri, AbcStream):
            uri.unlink(missing_ok=True, **kwargs)  # type: ignore
        else:
            self.open(uri).unlink(missing_ok=missing_ok, **kwargs)

    def rmdir(
        self,
        uri: Union[pathlib.Path, str, AbcStream],
        ignore_errors: bool = False,
        **kwargs,
    ):
        """Remove the entire directory."""
        if isinstance(uri, AbcStream):
            uri.rmdir(ignore_errors=ignore_errors, **kwargs)
        else:
            self.open(uri).rmdir(ignore_errors=ignore_errors, **kwargs)

    @classmethod
    def set_buffer_rollover_size(cls, value: int):
        """
        Set the max size of the buffer before the data is rollover to disk.

        Args:
            value (int): size before rollover
        """
        cls.INMEM_SIZE = value


stream_factory = Streams(*DEFAULT_STREAMS)
open_stream = stream_factory.open
Stream = stream_factory.open
register_stream = stream_factory.register
set_schema_kwargs = stream_factory.set_schema_kwargs
set_buffer_rollover_size = stream_factory.set_buffer_rollover_size
mkdir = stream_factory.mkdir
iter_dir = stream_factory.iter_dir
glob = stream_factory.glob
exists = stream_factory.exists
stats = stream_factory.stats
unlink = delete = remove = stream_factory.unlink
rmdir = stream_factory.rmdir



"""This module is for streams that requires optional dependencies - i.e. boto3 for S3Stream."""
import importlib
import importlib.util
import os

from iotoolz.extensions._not_implemented import mock_stream

BOTO3_EXIST = importlib.util.find_spec("boto3") is not None
S3Stream = (
    importlib.import_module("iotoolz.extensions.s3").S3Stream  # type: ignore
    if BOTO3_EXIST
    else mock_stream(
        supported_schemas={"s3", "s3a", "s3n"},
        msg="S3Stream is not available because 'boto3' is not installed. "
        "You can install 'boto3' yourself or use the command 'pip install iotoolz[boto3]' "
        "when installing iotoolz.",
    )
)



"""This module implements the TempStream using the original tempfile in AbcStream."""
import io
import os.path
import weakref
from typing import IO, Iterable, Iterator, Tuple, Union

from iotoolz._abc import AbcStream, StreamInfo
from iotoolz.utils import guess_content_type_from_buffer, guess_encoding

_TEMPSTREAMS: weakref.WeakValueDictionary = weakref.WeakValueDictionary()


class TempStream(AbcStream):
    """TempStream is the stream interface to an in-memory buffer with can rollover to local file system if the "inmem_size" arg is set."""

    supported_schemas = {"tmp", "temp"}

    def __init__(  # pylint: disable=too-many-arguments
        self,
        uri: str,
        data: Union[str, bytes, bytearray] = None,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = None,
        newline: str = None,
        content_type: str = "",
        inmem_size: int = None,
        delimiter: Union[str, bytes] = None,
        chunk_size: int = io.DEFAULT_BUFFER_SIZE,
        etag: str = "",
        **kwargs,
    ):
        """
        Creates a new instance of TempStream.

        Args:
            uri (str): uri string to the resource.
            data Union[str, bytes, bytearray]): initial data to the stream.
            mode (str, optional): same as "open" - supports depends on the actual implementation. Defaults to "r".
            buffering (int, optional): same as "open". Defaults to -1.
            encoding (str, optional): encoding used to decode bytes to str. Defaults to None.
            newline (str, optional): same as "open". Defaults to None.
            content_type (str, optional): mime type for the resource. Defaults to "".
            inmem_size (int, optional): max size before buffer rollover from mem to disk. Defaults to None (i.e. never - may raise MemoryError).
            delimiter (Union[str, bytes], optional): delimiter used for determining line boundaries. Defaults to None.
            chunk_size (int, optional): chunk size when iterating bytes stream. Defaults to io.DEFAULT_BUFFER_SIZE.
            etag (str, optional): etag for the stream content. Defaults to "".

        """
        super().__init__(
            uri,
            mode,
            buffering,
            encoding,
            newline,
            content_type,
            inmem_size,
            delimiter,
            chunk_size,
            etag,
            **kwargs,
        )
        self.set_info(content_type=content_type, encoding=encoding or "utf-8")
        self._initial_data = None
        self._has_read = True  # nothing to read
        self._has_stats = True

        if data:
            if "r" not in self.mode:
                self.mode = "r" + self.mode
                self.mode = "".join(
                    [char for char in self.mode if char not in {"w", "a"}]
                )
            if isinstance(data, (bytes, bytearray)):
                content_type = content_type or guess_content_type_from_buffer(data)
                encoding = encoding or guess_encoding([data])[0]
                if "b" not in self.mode:
                    self.mode += "b"
                self.set_info(content_type=content_type, encoding=encoding)
                self._file.write(data)
            elif isinstance(data, str):
                self.mode = "".join([char for char in self.mode if char != "b"])
                encoding = encoding or "utf-8"
                content_type = content_type or guess_content_type_from_buffer(
                    data.encode(encoding or "utf-8")
                )
                self.set_info(content_type=content_type, encoding=encoding)
                self._file.write(data.encode(encoding or "utf-8"))
            else:
                raise TypeError("data must be of type bytes or str")
            self._file.seek(0)

        # register to weakdict
        _TEMPSTREAMS[self.uri] = self

    def read_to_iterable_(
        self, uri: str, chunk_size: int, fileobj: IO[bytes], **kwargs
    ) -> Tuple[Iterable[bytes], StreamInfo]:
        return [], StreamInfo()

    def write_from_fileobj_(
        self, uri: str, fileobj: IO[bytes], size: int, **kwargs
    ) -> StreamInfo:
        return StreamInfo()

    def stats_(self) -> StreamInfo:
        return StreamInfo()

    def unlink(self, missing_ok: bool = True, **kwargs):
        self._clear_buffer()

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False,
    ):
        """This method does nothing as you do not need to create a folder for an in-memory buffer."""
        ...

    def iter_dir_(self) -> Iterable[StreamInfo]:
        dirpath = self.uri if self.uri.endswith("/") else os.path.dirname(self.uri)

        return (
            stream.info
            for uri, stream in _TEMPSTREAMS.items()
            if uri.startswith(dirpath)
            if _TEMPSTREAMS.get(uri)
        )

    def iter_dir(self) -> Iterator["TempStream"]:
        """
        If the current stream is a directory, this method will yield all Stream
        in the directory. Otherwise, it should yield all Stream in the same
        directory (or level) as the current stream.
        """
        dirpath = self.uri if self.uri.endswith("/") else os.path.dirname(self.uri)

        return (
            stream
            for uri, stream in _TEMPSTREAMS.items()
            if uri.startswith(dirpath)
            if _TEMPSTREAMS.get(uri)
        )

    def exists(self) -> bool:
        """TempStream will always exist."""
        return True

    @classmethod
    def open(
        cls,
        uri: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = None,
        newline: str = None,
        inmem_size: int = None,
        delimiter: Union[str, bytes] = None,
        chunk_size: int = io.DEFAULT_BUFFER_SIZE,
        etag: str = "",
        **kwargs,
    ) -> "TempStream":
        """
        Creates a new instance of TempStream.

        Args:
            uri (str): uri string to the resource.
            mode (str, optional): same as "open" - supports depends on the actual implementation. Defaults to "r".
            buffering (int, optional): same as "open". Defaults to -1.
            encoding (str, optional): encoding used to decode bytes to str. Defaults to None.
            newline (str, optional): same as "open". Defaults to None.
            content_type (str, optional): mime type for the resource. Defaults to "".
            inmem_size (int, optional): max size before buffer rollover from mem to disk. Defaults to None (i.e. never - may raise MemoryError).
            delimiter (Union[str, bytes], optional): delimiter used for determining line boundaries. Defaults to None.
            chunk_size (int, optional): chunk size when iterating bytes stream. Defaults to io.DEFAULT_BUFFER_SIZE.
            etag (str, optional): etag for the stream content. Defaults to "".
            **kwargs: Additional keyword arguments to the stream (depends on implementation)

        Returns:
            TempStream: new instance of TempStream
        """
        if uri in _TEMPSTREAMS:
            return _TEMPSTREAMS[uri]  # type: ignore
        return cls(
            uri,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            newline=newline,
            inmem_size=inmem_size,
            delimiter=delimiter,
            chunk_size=chunk_size,
            **kwargs,
        )



"""This module implements the TempStream using the original tempfile in AbcStream."""
from typing import IO, Iterable, Set, Tuple, Type

from iotoolz._abc import AbcStream, StreamInfo


def mock_stream(supported_schemas: Set[str], msg: str) -> Type[AbcStream]:
    class NotImplementedStream(AbcStream):
        """NotImplementedStream is a mock stream interface for extension streams that does not have the required dependencies installed."""

        def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called
            raise NotImplementedError(msg)

        def read_to_iterable_(
            self, uri: str, chunk_size: int, fileobj: IO[bytes], **kwargs
        ) -> Tuple[Iterable[bytes], StreamInfo]:
            raise NotImplementedError(msg)

        def write_from_fileobj_(
            self, uri: str, fileobj: IO[bytes], size: int, **kwargs
        ) -> StreamInfo:
            raise NotImplementedError(msg)

    NotImplementedStream.supported_schemas = supported_schemas

    return NotImplementedStream



"""This module implements the FileStream with python native "open" method."""
import datetime
import io
import os.path
import urllib.parse
from typing import IO, Any, Dict, Iterable, Tuple, Type, Union

try:
    import boto3
    import boto3.s3.transfer
    import botocore.errorfactory
except ImportError as error:
    raise ImportError(
        "S3Stream cannot be used because 'boto3' is not installed. "
        "You can install boto3 by running the command: 'pip install iotoolz[boto3]'"
    ) from error

import cytoolz

from iotoolz._abc import AbcStream, StreamInfo
from iotoolz.utils import guess_filename

ALLOWED_DOWNLOAD_ARGS = frozenset(boto3.s3.transfer.S3Transfer.ALLOWED_DOWNLOAD_ARGS)
ALLOWED_UPLOAD_ARGS = frozenset(boto3.s3.transfer.S3Transfer.ALLOWED_UPLOAD_ARGS)
ALLOWED_HEAD_ARGS = {
    "VersionId",
    "SSECustomerAlgorithm",
    "SSECustomerKey",
    "RequestPayer",
}
ALLOWED_DELETE_ARGS = {
    "VersionId",
    "RequestPayer",
    "BypassGovernanceRetention",
    "ExpectedBucketOwner",
}


class S3Stream(AbcStream):
    """S3Stream is the stream interface to AWS S3 object store.

    See https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.S3Transfer
    """

    supported_schemas = {"s3", "s3a", "s3n"}
    _default_client: boto3.client = None
    _default_transfer_config = boto3.s3.transfer.TransferConfig()
    _default_extra_upload_args: Dict[str, Any] = {}
    _default_extra_download_args: Dict[str, Any] = {}

    def __init__(
        self,
        uri: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = None,
        newline: str = None,
        content_type: str = "",
        inmem_size: int = None,
        delimiter: Union[str, bytes] = None,
        chunk_size: int = io.DEFAULT_BUFFER_SIZE,
        client: boto3.client = None,
        multipart_threshold: int = None,
        max_concurrency: int = None,
        multipart_chunksize: int = None,
        num_download_attempts: int = None,
        max_io_queue: int = None,
        io_chunksize: int = None,
        use_threads: bool = None,
        **kwargs,
    ):
        """
        Creates a new instance of S3Stream.

        See also: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.S3Transfer

        Args:
            uri (str): uri string to the resource.
            mode (str, optional): same as "open" - supports depends on the actual implementation. Defaults to "r".
            buffering (int, optional): same as "open". Defaults to -1.
            encoding (str, optional): encoding used to decode bytes to str. Defaults to None.
            newline (str, optional): same as "open". Defaults to None.
            content_type (str, optional): mime type for the resource. Defaults to "".
            inmem_size (int, optional): max size before buffer rollover from mem to disk. Defaults to None (i.e. never - may raise MemoryError).
            delimiter (Union[str, bytes], optional): delimiter used for determining line boundaries. Defaults to None.
            chunk_size (int, optional): chunk size when iterating bytes stream. Defaults to io.DEFAULT_BUFFER_SIZE.
            client (boto3.client, optional): use the provided boto3 client to interface with S3. Defaults to None.
            multipart_threshold (int, optional): The transfer size threshold for which multipart uploads, downloads, and copies will automatically be triggered. Defaults to 8388608.
            max_concurrency (int, optional): The maximum number of threads that will be making requests to perform a transfer. If use_threads is set to False, the value provided is ignored as the transfer will only ever use the main thread. Defaults to 10.
            multipart_chunksize (int, optional): The partition size of each part for a multipart transfer. Defaults to 8388608.
            num_download_attempts (int, optional): The number of download attempts that will be retried upon errors with downloading an object in S3. Note that these retries account for errors that occur when streaming down the data from s3 (i.e. socket errors and read timeouts that occur after receiving an OK response from s3). Other retryable exceptions such as throttling errors and 5xx errors are already retried by botocore (this default is 5). This does not take into account the number of exceptions retried by botocore. Defaults to 5.
            max_io_queue (int, optional): The maximum amount of read parts that can be queued in memory to be written for a download. The size of each of these read parts is at most the size of io_chunksize. Defaults to 100.
            io_chunksize (int, optional): The max size of each chunk in the io queue. Currently, this is size used when read is called on the downloaded stream as well. Defaults to 262144.
            use_threads (bool, optional): If True, threads will be used when performing S3 transfers. If False, no threads will be used in performing transfers: all logic will be ran in the main thread. Defaults to True.
            **kwargs: Additional ExtraArgs which will be passed to the 'boto3.s3.transfer.S3Transfer' client.
        """
        super().__init__(
            uri,
            mode,
            buffering,
            encoding,
            newline,
            content_type,
            inmem_size,
            delimiter,
            chunk_size,
            **kwargs,
        )
        self._client = client or self._default_client or boto3.client("s3")
        self._transfer_config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=multipart_threshold
            or self._default_transfer_config.multipart_threshold,
            max_concurrency=max_concurrency
            or self._default_transfer_config.max_request_concurrency,
            multipart_chunksize=multipart_chunksize
            or self._default_transfer_config.multipart_chunksize,
            num_download_attempts=num_download_attempts
            or self._default_transfer_config.num_download_attempts,
            max_io_queue=max_io_queue
            or self._default_transfer_config.max_io_queue_size,
            io_chunksize=io_chunksize or self._default_transfer_config.io_chunksize,
            use_threads=use_threads or self._default_transfer_config.use_threads,
        )
        self._scheme, self.bucket, self.key, query, _ = urllib.parse.urlsplit(uri)
        if self.key.startswith("/"):
            self.key = self.key[1:]
        # keep last query value
        query_str = {
            key: value[-1]
            for key, value in urllib.parse.parse_qs(query).items()
            if value
        }
        kwargs = {**kwargs, **query_str}
        extra_download_args = {
            key: value for key, value in kwargs.items() if key in ALLOWED_DOWNLOAD_ARGS
        }
        self._extra_download_args: Dict[str, Any] = {
            **self._default_extra_download_args,
            **extra_download_args,
        }
        extra_upload_args = {
            key: value for key, value in kwargs.items() if key in ALLOWED_UPLOAD_ARGS
        }
        self._extra_upload_args: Dict[str, Any] = {
            **self._default_extra_upload_args,
            **extra_upload_args,
        }
        self._head_args = {
            key: value for key, value in kwargs.items() if key in ALLOWED_HEAD_ARGS
        }
        self._delete_args = {
            key: value for key, value in kwargs.items() if key in ALLOWED_DELETE_ARGS
        }

    def read_to_iterable_(
        self, uri: str, chunk_size: int, fileobj: IO[bytes], **kwargs
    ) -> Tuple[Iterable[bytes], StreamInfo]:
        """Downloads the S3 object to buffer with 'boto3.s3.transfer.S3Transfer'."""
        resp = self._client.head_object(
            Bucket=self.bucket, Key=self.key, **self._extra_download_args
        )
        etag = resp.get("ETag", "").strip('"')
        encoding = resp.get("ContentEncoding", self.encoding)
        content_type = resp.get("ContentType", self.content_type)
        last_modified = resp.get("LastModified")

        self._client.download_fileobj(
            self.bucket,
            self.key,
            fileobj,
            ExtraArgs=self._extra_download_args,
            Config=self._transfer_config,
        )
        fileobj.seek(0)  # reset to initial counter
        return (
            [],
            StreamInfo(
                content_type=content_type,
                encoding=encoding,
                etag=etag,
                last_modified=last_modified,
                extras=resp,
            ),
        )

    def write_from_fileobj_(
        self, uri: str, fileobj: IO[bytes], size: int, **kwargs
    ) -> StreamInfo:
        """Uploads the data in the buffer with 'boto3.s3.transfer.S3Transfer'."""
        self._client.upload_fileobj(
            fileobj,
            self.bucket,
            self.key,
            ExtraArgs={
                "ContentType": self.content_type,
                "ContentEncoding": self.encoding,
                **self._extra_upload_args,
            },
            Config=self._transfer_config,
        )
        return StreamInfo(last_modified=datetime.datetime.now())

    def stats_(self) -> StreamInfo:
        resp = self._client.head_object(
            Bucket=self.bucket, Key=self.key, **self._head_args
        )
        return StreamInfo(
            content_type=resp.get("ContentType"),
            encoding=resp.get("ContentEncoding"),
            etag=resp.get("ETag", "").strip('"'),
            last_modified=resp.get("LastModified"),
            extras=resp,
        )

    def exists(self) -> bool:
        """Whether the stream points to an existing resource."""
        try:
            self.set_info(self.stats_())
            return True
        except botocore.errorfactory.ClientError:
            return False

    def is_dir(self) -> bool:
        """Whether stream points to a existing dir."""
        return self.uri.endswith("/")

    def is_file(self) -> bool:
        """Whether stream points to a existing file."""
        return self.exists()

    def unlink(self, missing_ok: bool = True, **kwargs):
        try:
            kwargs = {**self._delete_args, **kwargs}
            self._client.delete_object(
                Bucket=self.bucket, Key=self.key, **kwargs,
            )
        except botocore.errorfactory.ClientError:
            if not missing_ok:
                raise

    def rmdir(self, ignore_errors: bool = False, **kwargs) -> "S3Stream":
        """Remove the entire directory."""
        try:
            kwargs = {**self._delete_args, **kwargs}
            batched = cytoolz.partition_all(1000, self.iter_dir())
            for batch in batched:
                self._client.delete_objects(
                    Bucket=self.bucket,
                    Delete={"Objects": [{"Key": stream.key} for stream in batch]},
                    **kwargs,
                )
        except botocore.errorfactory.ClientError:
            if not ignore_errors:
                raise
        return self

    @classmethod
    def set_default_client(cls, client: boto3.client) -> Type["S3Stream"]:
        """
        Set the default boto3 client to use.

        Args:
            client (boto3.client): boto3 client object.
        """
        cls._default_client = client
        return cls

    @classmethod
    def set_default_transfer_config(
        cls,
        multipart_threshold: int = 8388608,
        max_concurrency: int = 10,
        multipart_chunksize: int = 8388608,
        num_download_attempts: int = 5,
        max_io_queue: int = 100,
        io_chunksize: int = 262144,
        use_threads: bool = True,
    ) -> Type["S3Stream"]:
        """
        Set the default transfer config.

        See also https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig

        Args:
            multipart_threshold (int, optional): The transfer size threshold for which multipart uploads, downloads, and copies will automatically be triggered. Defaults to 8388608.
            max_concurrency (int, optional): The maximum number of threads that will be making requests to perform a transfer. If use_threads is set to False, the value provided is ignored as the transfer will only ever use the main thread. Defaults to 10.
            multipart_chunksize (int, optional): The partition size of each part for a multipart transfer. Defaults to 8388608.
            num_download_attempts (int, optional): The number of download attempts that will be retried upon errors with downloading an object in S3. Note that these retries account for errors that occur when streaming down the data from s3 (i.e. socket errors and read timeouts that occur after receiving an OK response from s3). Other retryable exceptions such as throttling errors and 5xx errors are already retried by botocore (this default is 5). This does not take into account the number of exceptions retried by botocore. Defaults to 5.
            max_io_queue (int, optional): The maximum amount of read parts that can be queued in memory to be written for a download. The size of each of these read parts is at most the size of io_chunksize. Defaults to 100.
            io_chunksize (int, optional): The max size of each chunk in the io queue. Currently, this is size used when read is called on the downloaded stream as well. Defaults to 262144.
            use_threads (bool, optional): If True, threads will be used when performing S3 transfers. If False, no threads will be used in performing transfers: all logic will be ran in the main thread. Defaults to True.
        """

        cls._default_transfer_config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=multipart_threshold,
            max_concurrency=max_concurrency,
            multipart_chunksize=multipart_chunksize,
            num_download_attempts=num_download_attempts,
            max_io_queue=max_io_queue,
            io_chunksize=io_chunksize,
            use_threads=use_threads,
        )
        return cls

    @classmethod
    def set_default_download_args(cls, **kwargs) -> Type["S3Stream"]:
        """
        Set the default ExtraArgs to the 'boto3.s3.transfer.S3Transfer' download client.

        See also: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.S3Transfer

        The available kwargs are:

        'VersionId', 'SSECustomerAlgorithm', 'SSECustomerKey', 'SSECustomerKeyMD5',
        'RequestPayer'
        """
        cls._default_extra_download_args = {
            key: value for key, value in kwargs.items() if key in ALLOWED_DOWNLOAD_ARGS
        }
        return cls

    @classmethod
    def set_default_upload_args(cls, **kwargs) -> Type["S3Stream"]:
        """
        Set the default ExtraArgs to the 'boto3.s3.transfer.S3Transfer' upload client.

        See also: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.S3Transfer

        The available kwargs are:

        'ACL', 'CacheControl', 'ContentDisposition', 'ContentEncoding',
        'ContentLanguage', 'ContentType', 'Expires', 'GrantFullControl', 'GrantRead',
        'GrantReadACP', 'GrantWriteACP', 'Metadata', 'RequestPayer',
        'ServerSideEncryption', 'StorageClass', 'SSECustomerAlgorithm',
        'SSECustomerKey', 'SSECustomerKeyMD5', 'SSEKMSKeyId', 'Tagging',
        'WebsiteRedirectLocation'
        """
        cls._default_extra_upload_args = {
            key: value for key, value in kwargs.items() if key in ALLOWED_UPLOAD_ARGS
        }
        return cls

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False,
    ):
        """This method does nothing as you do not need to create a 'folder' for an object store."""
        ...

    def iter_dir_(self) -> Iterable[StreamInfo]:
        """Yields tuple of uri and the metadata in a directory."""
        continuation_token: str = ""
        if self.is_dir():
            prefix = self.key
        else:
            prefix = os.path.dirname(self.key)

        while True:
            list_kwargs = {"MaxKeys": 1000, "Prefix": prefix, "Bucket": self.bucket}
            if continuation_token:
                list_kwargs["ContinuationToken"] = continuation_token
            response = self._client.list_objects_v2(**list_kwargs)

            continuation_token = response.get("NextContinuationToken")
            for content in response.get("Contents", []):
                key = content.get("Key")
                if key:
                    etag = content.get("ETag", "").strip('"')
                    last_modified = content.get("LastModified")
                    uri = f"{self._scheme}://{self.bucket}/{key}"
                    yield StreamInfo(
                        uri=uri,
                        name=guess_filename(uri),
                        last_modified=last_modified,
                        etag=etag,
                        extras=content,
                    )
            if not response.get("IsTruncated"):  # At the end of the list?
                break



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iotoolz', 'iotoolz.extensions']

package_data = \
{'': ['*']}

install_requires = \
['cachetools>=4.1.1,<5.0.0',
 'chardet>=3.0.4,<4.0.0',
 'cytoolz>=0.11.0,<0.12.0',
 'pydantic>=0.30',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'requests>=2.24.0,<3.0.0',
 'toml>=0.10.1,<0.11.0',
 'typing-extensions']

extras_require = \
{':sys_platform == "darwin" or sys_platform == "windows"': ['python-magic-bin>=0.4.0,<0.5.0'],
 ':sys_platform == "linux"': ['python-magic>=0.4.0,<0.5.0'],
 'all': ['boto3>=1.16.5,<2.0.0', 'minio>=6.0.0,<7.0.0'],
 'boto3': ['boto3>=1.16.5,<2.0.0'],
 'minio': ['minio>=6.0.0,<7.0.0']}

setup_kwargs = {
    'name': 'iotoolz',
    'version': '0.1.0rc11',
    'description': 'Consistent io iterface to read and write from/to both local and different remote resources (e.g. http, s3)',
    'long_description': '# iotoolz\n\n[![PyPI version](https://badge.fury.io/py/iotoolz.svg)](https://badge.fury.io/py/iotoolz)\n[![Build Status](https://travis-ci.com/e2fyi/iotoolz.svg?branch=master)](https://travis-ci.com/github/e2fyi/iotoolz)\n[![Coverage Status](https://coveralls.io/repos/github/e2fyi/iotoolz/badge.svg?branch=master)](https://coveralls.io/github/e2fyi/iotoolz?branch=master)\n[![Documentation Status](https://readthedocs.org/projects/iotoolz/badge/?version=latest)](https://iotoolz.readthedocs.io/en/latest/?badge=latest)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Downloads](https://pepy.tech/badge/iotoolz/month)](https://pepy.tech/project/iotoolz)\n\n`iotoolz` is an improvement over `e2fyi-utils` and is inspired partly by `toolz`.\n`iotoolz` is a lib to help provide a consistent dev-x for interacting with any IO resources.\nIt provides an abstract class `iotoolz.AbcStream` which mimics python\'s native `open`\nvery closely (with some additional parameters and methods such as `save`).\n\nAPI documentation can be found at [https://iotoolz.readthedocs.io/en/latest/](https://iotoolz.readthedocs.io/en/latest/).\n\nChange logs are available in [CHANGELOG.md](https://github.com/e2fyi/iotoolz/blob/master/CHANGELOG.md).\n\n> - Python 3.6 and above\n> - Licensed under [Apache-2.0](./LICENSE).\n\n## Supported streams\n\nCurrent the following streams are supported:\n\n- `iotoolz.FileStream`: wrapper over built-in `open` function (`file://`)\n- `iotoolz.TempStream`: in-memory stream that will rollover to disk (`tmp://`, `temp://`)\n- `iotoolz.HttpStream`: http or https stream implemented with `requests` (`http://`, `https://`)\n- `iotoolz.extensions.S3Stream`: s3 stream implemented with `boto3` (`s3://`, `s3a://`, `s3n://`)\n\n## Installation\n\n```bash\n# install the default packages only (most lite-weight)\npip install iotoolz\n\n# install dependencies for specific extension\npip install iotoolz[boto3]\n\n# install all the extras\npip install iotoolz[all]\n```\n\nAvailable extras:\n\n- `all`: All the optional dependencies\n- `boto3`: `boto3` for `iotoolz.extensions.S3Stream`\n- `minio`: TODO\n\n## Quickstart\n\nThe helper object `iotoolz.streams.stream_factory` is a default singleton of\n`iotoolz.streams.Streams` provided to support most of the common use cases.\n\n`iotoolz.streams.open_stream` (alias `iotoolz.streams.Stream`) and is a util method\nprovided by the singleton helper to create a stream object. This method accepts the same\narguments as python\'s `open` method with the following additional parameters:\n\n- `data`: optional str or bytes that will be passed into the stream\n- `fileobj`: optional file-like object which will be copied into the stream\n- `content_type`: optional mime type information to describe the stream (e.g. application/json)\n- `inmem_size`: determines how much memory to allocate to the stream before rolling over to local file system. Defaults to no limits (may result in MemoryError).\n- `schema_kwargs`: optional mapping of schemas to their default kwargs.\n\n### Basic Setup\n\n```py\nfrom iotoolz.streams import (\n    set_schema_kwargs,\n    set_buffer_rollover_size,\n)\n\n# set params to pass to the Stream obj handling https\n# i.e. HttpStream (implemented with requests)\nset_schema_kwargs(\n    "https",\n    verify=False,  # do not verify ssl cert\n    use_post=True  # use POST instead of PUT when writing to https\n)\n\n# use a custom client for S3Stream (via boto3)\nset_schema_kwargs(\n    "s3",\n    client=boto3.client(\n        "s3",\n        aws_access_key_id=ACCESS_KEY,\n        aws_secret_access_key=SECRET_KEY,\n        aws_session_token=SESSION_TOKEN,\n    )\n)\n\n# buffer will rollover to disk if the data is more than 100 MB\n# (default is everything is in-memory - may result in memory error)\nset_buffer_rollover_size(10**8)\n```\n\n### Opening streams\n\nYou can open any stream just like python\'s built-in `open` method.\n\n```py\nimport pandas\n\nfrom iotoolz import open_stream\n\n# print line by line some data in from a https endpoint\n# and do not verify the ssl cert of the https endpoint\nwith open_stream(\n    "https://foo/bar/data.txt",\n    mode="r",\n    schema_kwargs={"https": {"verify": False}}\n) as stream:\n    for line in stream:\n        print(line)\n\n# POST some binary content to a http endpoint (default is PUT)\nwith open_stream("https://foo.bar/api/data", "wb", use_post=True) as stream:\n    stream.write(b"hello world")\n\n# Copying a local file to s3\nwith open_stream("path/to/data.csv", "r") as csv_source,\n     open_stream("s3://bucket/foobar.txt?StorageClass=STANDARD", "w") as s3_sink:\n    # pipe content in csv_source to tmpsink\n    csv_source.pipe(s3_sink)\n\n# load to pandas dataframe from s3 fileobj\nwith open_stream("s3://bucket/foobar.csv", "r") as csv:\n    df = pd.read_csv(csv)\n\n```\n\n## TempStream\n\n`TempStream` is a stream can functions like a virtual file system in memory.\n\n```py\nimport gc\n\nfrom iotoolz import Stream, exists, glob, iter_dir\n\n# this stream can be garbage collected\nStream("tmp://foo/bar/data.txt", data="foobar")\n\n# True if not gc yet, False if already gc\nexists("tmp://foo/bar/data.txt")\n\n# force gc\ngc.collect()\n# will not exist\nexists("tmp://foo/bar/data.txt")\n\n# create temp stream with strong ref (hence will not be gc)\ns1 = Stream("tmp://foo/bar/data.txt", data="foobar")\ns2 = Stream("tmp://foo/example.txt", data="...")\n\n# returns s1 and s2\niter_dir("tmp://foo/")\n\n# returns s1 only\nglob("tmp://foo/bar/*.txt")\n```\n\n## Stream-like operations\n\n`Stream` is an alias of `open_stream`, both methods return a concrete `AbcStream` object.\nYou can treat the object as both a "file-like" and "stream-like" object - i.e. you can\nread, write, seek, flush, close the object.\n\n> NOTE\n>\n> By default, the underlying buffer is in-memory. You can enable rollover to disk by\n> passing the `inmem_size` arg to the method, or update the default `inmem_size` value\n> with the `iotoolz.streams.set_buffer_rollover_size` method.\n\n```py\nfrom iotoolz import open_stream, Stream, set_buffer_rollover_size\n\n# `Stream` is an alias of `open_stream`\nassert open_stream == Stream\n\n# rollover to disk if data is over 100 MB\nset_buffer_rollover_size(10**8)\n\n# you can overwrite the default kwargs here also\nstream = Stream(\n    "path/to/data.txt",\n    mode="rw",  # you can both read and write to a stream\n)\n# stream is lazily evaluated, nothing will be buffered until you call some methods\n# that requires the data\ndata = stream.read()\n# will attempt to provide encoding and content_type if not provided when opening the stream\nprint(stream.encoding)\nprint(stream.content_type)\n# stream has the same interface as an IO object - i.e. u can seek, flush, close, etc\nstream.seek(5)  # go to offset 5 from start of buffer\nstream.write("replace with this text")\nstream.seek(0, whence=2)  # go to end of buffer\nstream.write("additional text after original eof")  # continue writing to the end of the buffer\nstream.save()  # flush save the entire buffer to the same dst location\nstream.close() # close the stream\n```\n\n## Path-like operations\n\n`exists`, `mkdir`, `iter_dir` and `glob` are path-like methods that are available to the\nstream object. These methods mimics their equivalent in `pathlib.Path` when appropriate.\n\n| method                       | supported streams                           | desc                                                                                       |\n| ---------------------------- | ------------------------------------------- | ------------------------------------------------------------------------------------------ |\n| `stats`                      | All Streams                                 | return the StreamInfo for an existing resource                                             |\n| `unlink`, `delete`, `remove` | All Streams                                 | Delete and remove the stream (except for `TempStream` where the buffer is cleared instead) |\n| `exists`                     | All Streams                                 | check if a stream points to an existing resource.                                          |\n| `mkdir`                      | `FileStream`                                | create a directory.                                                                        |\n| `rmdir`                      | `FileStream`, `TempStream`, and `S3Stream`, | remove recursively everything in the directory.                                            |\n| `iter_dir`                   | `FileStream`, `TempStream`, and `S3Stream`  | iterate thru the streams in the directory.                                                 |\n| `glob`                       | `FileStream`, `TempStream`, and `S3Stream`  | iterate thru the streams in the directory that match a pattern.                            |\n\n```py\nimport itertools\n\nfrom iotoolz import Stream, mkdir, iter_dir, glob, exists\n\n# similar to \'mkdir -p\'\nmkdir("path/to/folder", parents=True, exist_ok=True)\nStream("path/to/folder").mkdir(parents=True, exist_ok=True)\n\n# list object in an s3 bucket\niter_dir("s3://bucket/prefix/")\nfor stream in Stream("s3://bucket/prefix/").iter_dir():\n    print(stream.uri)\n\n# find s3 objects with a specific pattern\nglob("s3://bucket/prefix/*txt")\nfor stream in Stream("s3://bucket/prefix/").glob("*.txt"):\n    print(stream.uri)\n\n# exists\nexists("s3://bucket/prefix/foo.txt")\n\n# stats\ninfo = stats("s3://bucket/prefix/foo.txt")\nprint(info.name)\nprint(info.content_type)\nprint(info.encoding)\nprint(info.last_modified)\nprint(info.etag)\nprint(info.extras)\n\n# delete resource\nunlink("s3://bucket/prefix/foo.txt")\n\n# rm all key with prefix\nrmdir("s3://bucket/prefix/")\n\n```\n\n## Piping streams\n\n`pipe` is method to push data to a sink (similar to NodeJS stream except it has no\nwatermark or buffering).\n\n```py\nfrom  iotoolz.streams import open_stream\n\nlocal_file = open_stream(\n    "path/to/google.html", content_type="text/html", mode="w"\n)\ntemp_file = open_stream(\n    "tmp://google.html", content_type="text/html", mode="wb"\n)\n\n# when source is closed, all sinks will be closed also\nwith open_stream("https://google.com") as source:\n    # writes to a temp file then to a local file in sequence\n    source.pipe(temp_file).pipe(local_file)\n\n\nlocal_file2 = open_stream(\n    "path/to/google1.html", content_type="text/html", mode="w"\n)\nlocal_file3 = open_stream(\n    "path/to/google2.html", content_type="text/html", mode="w"\n)\n\n# when source is closed, all sinks will be closed also\nwith open_stream("tmp://foo_src", mode="w") as source:\n    # writes in a fan shape manner\n    source.pipe(local_file2)\n    source.pipe(local_file3)\n\n    source.write("hello world")\n```\n\n> TODO support transform streams so that pipe can be more useful\n',
    'author': 'eterna2',
    'author_email': 'eterna2@hotmail.com',
    'maintainer': 'eterna2',
    'maintainer_email': 'eterna2@hotmail.com',
    'url': 'https://github.com/e2fyi/iotoolz',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)



"""
This module provides the abstract class for a generic IOStream.
"""
import abc
import atexit
import dataclasses
import datetime
import fnmatch
import functools
import io
import os.path
import signal
import tempfile
import threading
import warnings
import weakref
from typing import (
    IO,
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from iotoolz.utils import (
    guess_content_type_from_buffer,
    guess_encoding,
    guess_filename,
    peek_stream,
)

T = TypeVar("T")


def _set_read_flag(method: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(method)
    def _wrapped_set_read_flag(self: "AbcStream", *args, **kwargs) -> Any:
        ret = method(self, *args, **kwargs)
        self._has_read = True  # pylint: disable=protected-access
        return ret

    return _wrapped_set_read_flag


def need_sync(method: Callable[..., Any]) -> Callable[..., Any]:
    """
    need_sync is a decorator function to apply on any AbcStream method.

    Any AbcStream methods that are decorated will always attempt to read from
    the resource source into the tempfile, before executing the method.

    Args:
        method ([Callable[..., Any]]): Any AbcStream method
    """

    @functools.wraps(method)
    def _implement_sync_fileobj(self: "AbcStream", *args, **kwargs) -> Any:
        self.sync()
        return method(self, *args, **kwargs)

    return _implement_sync_fileobj


@dataclasses.dataclass
class StreamInfo:
    """
    dataclass to wrap around some basic info about the stream.
    """

    uri: str = ""
    name: str = ""
    content_type: Optional[str] = ""
    encoding: Optional[str] = ""
    etag: Optional[str] = ""
    last_modified: Optional[datetime.datetime] = None
    extras: dict = dataclasses.field(default_factory=dict)


def merge_streaminfo(info1: StreamInfo, info2: StreamInfo) -> StreamInfo:
    return StreamInfo(
        uri=info2.uri or info1.uri,
        name=info2.name or info1.name,
        content_type=info2.content_type or info1.content_type,
        encoding=info2.encoding or info1.encoding,
        etag=info2.etag or info1.etag,
        last_modified=info2.last_modified or info1.last_modified,
        extras={**info1.extras, **info2.extras},
    )


def transform_streaminfo(info: StreamInfo, **kwargs) -> StreamInfo:
    kwargs = {**dataclasses.asdict(info), **kwargs}
    return StreamInfo(**kwargs)


class AbcStream(
    abc.ABC
):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """
    AbcStream is an abstract class which mimics python's native `open` function very
    closely.
    """

    INMEM_SIZE: int = 0
    supported_schemas: Set[str] = set()

    def __init__(
        self,
        uri: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = None,
        newline: str = None,
        content_type: str = "",
        inmem_size: int = None,
        delimiter: Union[str, bytes] = None,
        chunk_size: int = io.DEFAULT_BUFFER_SIZE,
        etag: str = "",
        **kwargs,
    ):
        """
        Creates a new instance of AbcStream.

        Args:
            uri (str): uri string to the resource.
            mode (str, optional): same as "open" - supports depends on the actual implementation. Defaults to "r".
            buffering (int, optional): same as "open". Defaults to -1.
            encoding (str, optional): encoding used to decode bytes to str. Defaults to None.
            newline (str, optional): same as "open". Defaults to None.
            content_type (str, optional): mime type for the resource. Defaults to "".
            inmem_size (int, optional): max size before buffer rollover from mem to disk. Defaults to None (i.e. never - may raise MemoryError).
            delimiter (Union[str, bytes], optional): delimiter used for determining line boundaries. Defaults to None.
            chunk_size (int, optional): chunk size when iterating bytes stream. Defaults to io.DEFAULT_BUFFER_SIZE.
            etag (str, optional): etag for the stream content. Defaults to "".
            **kwargs: Additional keyword arguments to the stream (depends on implementation)
        """
        self.uri = uri
        self.filename = os.path.basename(uri)
        self.mode = mode
        self.buffering = buffering
        self.newline = newline

        self._delimiter = delimiter
        self._chunk_size = chunk_size
        self._inmem_size = inmem_size
        self._kwargs = kwargs

        self._has_read = False
        self._has_stats = False
        self._buffer: Optional[IO[bytes]] = None
        self._info = StreamInfo(
            uri=uri,
            name=guess_filename(uri),
            content_type=content_type,
            encoding=encoding or "utf-8",
            etag=etag,
        )
        self._pipes: List[Tuple[str, IO]] = []

        _teardown = weakref.WeakMethod(self._teardown)  # type: ignore
        if threading.current_thread() is threading.main_thread():
            self._sigterm_handler = signal.signal(signal.SIGTERM, _teardown)  # type: ignore
            self._sigint_handler = signal.signal(signal.SIGINT, _teardown)  # type: ignore
            self._sigquit_handler = signal.signal(signal.SIGQUIT, _teardown)  # type: ignore
        else:
            self._sigterm_handler = self._sigint_handler = self._sigquit_handler = None

        self._stream_params = {
            "mode": mode,
            "buffering": buffering,
            "encoding": encoding,
            "newline": newline,
            "content_type": content_type,
            "inmem_size": inmem_size,
            "delimiter": delimiter,
            "chunk_size": chunk_size,
            **kwargs,
        }
        atexit.register(weakref.WeakMethod(self._cleanup))  # type: ignore

        # alias
        self.delete = self.unlink
        self.remove = self.unlink

    @abc.abstractmethod
    def read_to_iterable_(
        self, uri: str, chunk_size: int, fileobj: IO[bytes], **kwargs
    ) -> Tuple[Iterable[bytes], StreamInfo]:
        """
        read_to_iterable_ is an abstract method to implement the reading of the source
        resource into the a binary iterator - i.e. you will need to encode your data
        appropriately if it is a string.

        Alternatively, 'fileobj' argument is also provided where you can write to the
        stream file buffer directly. In this case, you should return an empty iterable.
        However, this will be less efficient as the actual read will only start after
        all the data have been read  into the buffer.

        The method should return a tuple of the bytes iterator and StreamInfo object
        which ideally should provide the following info:

        - content_type: (see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type)

        - encoding: (see https://docs.python.org/2.4/lib/standard-encodings.html)

        - etag: (see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/ETag)

        An optional dict "extras" is also provided for any other information.

        Args:
            uri (str): source uri to the resource
            chunk_size (int): size for each chunk
            fileobj: (IO[bytes]): temp fileobj for the stream

        Raises:
            NotImplementedError: [description]

        Returns:
            Tuple[Iterable[bytes], StreamInfo]: tuple of the iterable and StreamInfo
                object describing the data
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write_from_fileobj_(
        self, uri: str, fileobj: IO[bytes], size: int, **kwargs
    ) -> StreamInfo:
        """
        Abstract method to implement the write to the
        destination resource from the provided file-like object.

        This file-like object only provides binary outputs - i.e. if the resources only
        accepts text input, you will need to decode the output inside this method.

        Args:
            uri (str): destination uri of the resource
            fileobj (IO[bytes]): file-like object to read from
            size (int): size of the data inside the file-like object

        Raises:
            NotImplementedError: [description]

        Returns:
            StreamInfo: StreamInfo object describing the data
        """
        raise NotImplementedError

    @abc.abstractmethod
    def stats_(self) -> StreamInfo:
        """Retrieve the StreamInfo."""
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self) -> bool:
        """Whether the path points to an existing resource."""
        raise NotImplementedError

    @abc.abstractmethod
    def unlink(self, missing_ok: bool = True, **kwargs):
        """Delete and remove the resource."""
        raise NotImplementedError

    @abc.abstractmethod
    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False,
    ):
        """
        This abstract method should mimics pathlib.mkdir method for different streams.

        Create a new directory at this given path. If mode is given, it is combined with
        the process’ umask value to determine the file mode and access flags. If the path
        already exists, FileExistsError is raised.

        If parents is true, any missing parents of this path are created as needed; they
        are created with the default permissions without taking mode into account
        (mimicking the POSIX mkdir -p command).

        If parents is false (the default), a missing parent raises FileNotFoundError.

        If exist_ok is false (the default), FileExistsError is raised if the target
        directory already exists.

        If exist_ok is true, FileExistsError exceptions will be ignored (same behavior
        as the POSIX mkdir -p command), but only if the last path component is not an
        existing non-directory file.

        Args:
            mode (int, optional): mask mode. Defaults to 0o777.
            parents (bool, optional): If true, creates any parents if required. Defaults to False.
            exist_ok (bool, optional): If true, will not raise exception if dir already exists. Defaults to False.

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    @abc.abstractmethod
    def iter_dir_(self) -> Iterable[StreamInfo]:
        """
        If the current stream is a directory, this method should yield StreamInfo in
        the directory. Otherwise, it should yield other StreamInfo in the same
        directory (or level) as the current stream.
        """
        raise NotImplementedError

    @classmethod
    def open(
        cls,
        uri: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = None,
        newline: str = None,
        inmem_size: int = None,
        delimiter: Union[str, bytes] = None,
        chunk_size: int = io.DEFAULT_BUFFER_SIZE,
        etag: str = "",
        **kwargs,
    ) -> "AbcStream":
        """
        Creates a new instance of AbcStream. This mimics python's `open` method but
        supports additional keyword arguments.

        Args:
            uri (str): uri string to the resource.
            mode (str, optional): same as "open" - supports depends on the actual implementation. Defaults to "r".
            buffering (int, optional): same as "open". Defaults to -1.
            encoding (str, optional): encoding used to decode bytes to str. Defaults to None.
            newline (str, optional): same as "open". Defaults to None.
            content_type (str, optional): mime type for the resource. Defaults to "".
            inmem_size (int, optional): max size before buffer rollover from mem to disk. Defaults to None (i.e. never - may raise MemoryError).
            delimiter (Union[str, bytes], optional): delimiter used for determining line boundaries. Defaults to None.
            chunk_size (int, optional): chunk size when iterating bytes stream. Defaults to io.DEFAULT_BUFFER_SIZE.
            etag (str, optional): etag for the stream content. Defaults to "".
            **kwargs: Additional keyword arguments to the stream (depends on implementation)

        Returns:
            AbcStream: new instance of AbcStream
        """
        return cls(
            uri,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            newline=newline,
            inmem_size=inmem_size,
            delimiter=delimiter,
            chunk_size=chunk_size,
            etag=etag,
            **kwargs,
        )

    @property
    def _file(self) -> IO[bytes]:
        """Tempfile object for the stream."""
        return self._buffer or self._clear_buffer()

    @property
    def closed(self) -> bool:
        """True if the stream is closed."""
        return self._file.closed

    @property
    def encoding(self) -> str:
        """
        Text encoding to use to decode internal binary data into text outputs.

        Returns:
            str: text encoding.
        """
        if not self._info.encoding:
            encoding = guess_encoding(self.iter_bytes())[0]
            self.set_info(encoding=encoding)
        return self._info.encoding or ""

    @property
    def content_type(self) -> str:
        """
        Resource media type (e.g. application/json, text/plain).

        See also https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type

        Returns:
            str: string describing the media type of the resource.
        """
        if not self._info.content_type and self.size > 0:
            with peek_stream(self._file, peek=0) as stream:
                content_type = guess_content_type_from_buffer(stream.read(1024))  # type: ignore
            self.set_info(content_type=content_type)
        return self._info.content_type or ""

    @property
    def etag(self) -> str:
        """Identifier for a specific version of a resource."""
        return self._info.etag or ""

    @property
    def size(self) -> int:
        curr = self.tell()
        self.seek(0, 2)
        size = self.tell()
        self.seek(curr)
        return size

    @property
    def info(self) -> StreamInfo:
        """
        Stream info like content type, encoding, and etag.

        Returns:
            StreamInfo: StreamInfo object for the current stream.
        """
        if not self._has_stats and "r" in self.mode:
            return self.stats()
        return self._info

    def stats(self) -> StreamInfo:
        """Retrieve the StreamInfo of the current stream."""
        try:
            self.set_info(self.stats_())
            self._has_stats = True
        except Exception as error:  # pylint: disable=broad-except
            warnings.warn(f"{self.uri}: {error}", RuntimeWarning)
        return self._info

    def is_dir(self) -> bool:
        """Whether stream points to a existing dir."""
        return self.exists() and self.uri.endswith("/")

    def is_file(self) -> bool:
        """Whether stream points to a existing file."""
        return self.exists() and not self.uri.endswith("/")

    def rmdir(self, ignore_errors: bool = False, **kwargs) -> "AbcStream":
        """Remove the entire directory."""
        for stream in self.iter_dir():
            stream.unlink(missing_ok=ignore_errors, **kwargs)
        return self

    @need_sync
    def peek(self, size: Optional[int] = None) -> bytes:
        """
        Return bytes from the stream without advancing the position.

        Args:
            size (Optional[int], optional): number of bytes to return. Return default
            chunk size if not provided. Return everything if is -1. Defaults to None.

        Returns:
            bytes: data starting from current stream pos.
        """
        with peek_stream(self) as stream:
            return stream.read(size or self._chunk_size)  # type: ignore

    @need_sync
    def read(self, size: Optional[int] = -1) -> Union[str, bytes, bytearray]:
        """
        Read and return up to size bytes. If the argument is omitted, None, or negative,
        data is read and returned until EOF is reached. An empty bytes object is
        returned if the stream is already at EOF.

        Args:
            size (Optional[int], optional): number of bytes to read. Defaults to -1.

        Returns:
            Union[str, bytes, bytearray]: data starting from current stream pos.
        """
        chunk = self._file.read(size or -1)
        if "b" in self.mode:
            return chunk
        return chunk.decode(self.encoding)

    @need_sync
    def readline(self, size: Optional[int] = -1) -> Union[str, bytes, bytearray]:
        """
        Read and return one line from the stream. If size is specified, at most size
        bytes will be read.

        Args:
            size (Optional[int], optional): [description]. Defaults to -1.

        Returns:
            Union[str, bytes, bytearray]: [description]
        """
        chunk = self._file.readline(size or -1)
        if "b" in self.mode:
            return chunk
        return chunk.decode(self.encoding)

    @need_sync
    def readlines(
        self, hint: Optional[int] = -1
    ) -> Union[List[str], List[bytes], List[bytearray]]:
        """
        Read and return a list of lines from the stream. hint can be specified to
        control the number of lines read: no more lines will be read if the total size
        (in bytes/characters) of all lines so far exceeds hint.

        Args:
            hint (Optional[int], optional): [description]. Defaults to -1.

        Returns:
            Union[List[str], List[bytes], List[bytearray]]: [description]
        """
        chunks = self._file.readlines(hint or -1)
        if "b" in self.mode:
            return chunks
        return [chunk.decode(self.encoding) for chunk in chunks]

    def write(self, data: Union[str, bytes, bytearray]) -> int:
        """
        Write the given bytes-like object into buffer and return the number of bytes
        written.

        The data will not be written to the actual resource until "save" or "close"
        method is called.

        Args:
            data (Union[str, bytes, bytearray]): bytes-like object to write.

        Raises:
            IOError: stream is readonly.
            TypeError: expect data to be of type str, bytes or bytearray.

        Returns:
            int: number of bytes written.
        """
        if "r" in self.mode and "w" not in self.mode and "a" not in self.mode:
            raise IOError(f"{self} is readonly: {self.mode}")
        if isinstance(data, str):
            return self._write(data.encode(self.encoding))
        if not isinstance(data, (bytes, bytearray)):
            raise TypeError("expect data to be of type 'bytes' or 'bytearray'")
        return self._write(data)

    def seek(self, offset, whence: int = 0) -> int:
        """
        Change the stream position to the given byte offset. offset is interpreted
        relative to the position indicated by whence. The default value for whence is
        SEEK_SET. Values for whence are:

        SEEK_SET or 0 – start of the stream (the default); offset should be zero or positive

        SEEK_CUR or 1 – current stream position; offset may be negative

        SEEK_END or 2 – end of the stream; offset is usually negative

        Return the new absolute position.

        Args:
            offset ([type]): offset relative to position indicated by "whence".
            whence (int, optional): reference position - 0, 1, or 2. Defaults to 0.

        Returns:
            [int]: new absolute position.
        """
        return self._file.seek(offset, whence)

    def tell(self) -> int:
        """
        Return the current stream position.

        Returns:
            int: current stream position.
        """
        return self._file.tell()

    def flush(self):
        """
        Flush the write buffers of the stream if applicable.
        This does nothing for read-only and non-blocking streams.
        """
        self._file.flush()

    def save(
        self, data: Union[bytes, bytearray, str] = None, close: bool = False
    ) -> "AbcStream":
        """
        Flush and stream everything in the buffer to the actual resource location.

        Does nothing if mode is read-only. Will not close the stream.

        Args:
            data: (Union[bytes, bytearray, str], optional): Write additional data to stream before saving. Defaults to None.
            close (bool, optional): Close stream after saving. Defaults to False.

        Returns:
            [AbcStream]: current stream object.
        """
        if self.closed:
            return self

        if self.is_empty() and not data:
            if close:
                self.close()
            return self

        if "w" in self.mode or "a" in self.mode:
            if data:
                self.write(data)
            self.flush()
            with peek_stream(self._file, peek=0, ignore_closed=True) as stream:
                info = self.write_from_fileobj_(
                    self.uri, stream, self.size, **self._kwargs
                )
            self.set_info(info)

        if close:
            self.close()

        return self

    def close(self):
        """
        Flush and close this stream. This method has no effect if the file is already
        closed. Once the file is closed, any operation on the file (e.g. reading or
        writing) will raise a ValueError.

        As a convenience, it is allowed to call this method more than once; only the
        first call, however, will have an effect.
        """
        try:
            self.save()
            self._file.close()
            for _, sink in self._pipes:
                sink.close()
        finally:
            self._cleanup()

    def is_empty(self) -> bool:
        """
        True if current buffer is empty.

        Returns:
            bool: whether current buffer is empty.
        """
        # remember pos
        pos = self.tell()
        # go to the end of stream
        self.seek(0, 2)
        is_empty = self.tell() == 0
        # restore org pos
        self.seek(pos)
        return is_empty

    def set_info(self, info: StreamInfo = None, **kwargs) -> "AbcStream":
        if info:
            self._info = merge_streaminfo(self._info, info)
        if kwargs:
            self._info = transform_streaminfo(self._info, **kwargs)
        return self

    def set_encoding(self, encoding: str) -> "AbcStream":
        """
        Set and update the encoding to use to decode the binary data into text.

        Args:
            encoding ([str]): text encoding for the stream.

        Returns:
            AbcStream: current stream object.
        """
        self.set_info(encoding=encoding)
        for _, sink in self._pipes:
            if hasattr(sink, "set_encoding"):
                sink.set_encoding(encoding)  # type: ignore
        return self

    @_set_read_flag
    def _read_to_buffer(self) -> "AbcStream":
        iter_bytes, info = self.read_to_iterable_(
            self.uri, self._chunk_size, self._clear_buffer(), **self._kwargs
        )
        self.set_info(info, encoding=self.encoding, content_type=self.content_type)
        iter_bytes = iter_bytes or []
        for chunk in iter_bytes:
            self._write(chunk)
        self._file.seek(0)
        return self

    def sync(self, force: bool = False) -> "AbcStream":
        """
        Loads the resource into buffer if not loaded.

        Args:
            force ([bool]): If true, load the resource into buffer again even if it is already loaded.

        Returns:
            AbcStream: current stream object.
        """
        if not self._has_read or force:
            self._read_to_buffer()
        return self

    def clone(self, uri: str = None, info: StreamInfo = None, **kwargs) -> "AbcStream":
        """
        Creates a new stream with the same init args as the current stream.

        Returns:
            AbcStream: new stream with same init args.
        """
        kwargs = {**self._stream_params, **kwargs}
        obj = type(self)(uri or self.uri, **kwargs)
        if info:
            obj.set_info(info)  # pylint: disable=protected-access
        return obj

    def iter_dir(self) -> Iterator["AbcStream"]:
        """
        If the current stream is a directory, this method will yield all Stream
        in the directory. Otherwise, it should yield all Stream in the same
        directory (or level) as the current stream.
        """
        mode = self.mode.replace("w", "")
        if "r" not in mode:
            mode = f"r{mode}"
        # do not copy the content_type and encoding
        return (self.clone(info.uri, mode=mode, info=info) for info in self.iter_dir_())

    @_set_read_flag
    def _iter_bytes_from_reader(self) -> Iterator[bytes]:
        iter_bytes, info = self.read_to_iterable_(
            self.uri, self._chunk_size, self._file, **self._kwargs
        )
        self.set_info(info)
        self._clear_buffer()
        if iter_bytes:
            for chunk in iter_bytes:
                self._write(chunk)
                yield chunk
            self.seek(0)
        else:
            yield from self._iter_bytes_from_buffer()

    def _iter_bytes_from_buffer(self) -> Iterator[bytes]:
        with peek_stream(self._file, peek=0) as file_:
            read = functools.partial(file_.read, self._chunk_size)
            yield from iter(read, b"")
        self.seek(0)

    def iter_bytes(self) -> Iterator[bytes]:
        """
        Returns an iterator which yields a bytes stream. Will load the resource into
        buffer if needed.

        Yields:
            Iterator[bytes]: bytes stream from the start position
        """
        if not self._has_read:
            return self._iter_bytes_from_reader()  # type: ignore
        return self._iter_bytes_from_buffer()

    @need_sync
    def pipe(self, sink: IO, text_mode: bool = False) -> IO:
        """
        Pipes the data from the current stream object into any file-like object.

        Args:
            sink (IO): Any file-like object or AbcStream object.
            text_mode (bool, optional): If True, writes string to sink rather than bytes. Defaults to False.

        Raises:
            ValueError:  sink for pipe must be a filelike object - i.e. has write method

        Returns:
            IO: file-like object (i.e. sink) that is piped into.
        """
        if hasattr(sink, "write") and callable(sink.write):  # type: ignore
            # if empty, don't decode
            encoding = self.encoding or "utf-8"
            # update the encoding of the AbcStream to be same as source
            if hasattr(sink, "set_encoding") and encoding:
                sink.set_encoding(encoding)  # type: ignore
            self._pipes.append((encoding if text_mode else "", sink))
            # remember current pos
            pos = self.tell()
            # go to end of stream
            self._file.seek(0, 2)
            end = self.tell()
            # have some content
            if end > 0:
                # go to start
                self._file.seek(0)
                # stream existing to sink
                if text_mode:
                    for line in iter(self._file.readline, b""):
                        sink.write(line.decode(encoding))
                else:
                    for chunk in self._file:
                        sink.write(chunk)
            # go back to original position
            self.seek(pos)
            return sink

        raise ValueError(
            "sink for pipe must be a filelike object - i.e. has write method"
        )

    def glob(self, pattern: str = "*") -> Iterator["AbcStream"]:
        """
        Glob the given relative pattern in the directory represented by this path,
        yielding all matching files (of any kind).

        Args:
            pattern (str, optional): unix shell style pattern. Defaults to "*".

        Returns:
            Iterator["AbcStream"]: iterator of streams relative to current stream

        Yields:
            AbcStream: stream object
        """
        if self.is_dir():
            pattern = os.path.join(self.uri, pattern)
        else:
            pattern = os.path.join(os.path.dirname(self.uri), pattern)
        return (
            stream for stream in self.iter_dir() if fnmatch.fnmatch(stream.uri, pattern)
        )

    def _emit(self, chunk: bytes, pipes: List[Tuple[str, IO]]):
        for encoding, sink in pipes:
            data = chunk.decode(self.encoding or encoding) if encoding else chunk
            sink.write(data)

    def __iter__(self) -> Iterator[Union[bytes, str]]:
        """
        Yields chunks of bytes or lines depending on the mode.

        Yields:
            Iterator[Union[bytes, str]]: bytes stream if in binary mode or line stream for text mode.
        """
        # each iter will start from beginning
        self.seek(0)
        # for binary mode, can lazily write to temp and yield the chunk at the same time
        if "b" in self.mode:
            return self.iter_bytes()
        # for text mode, must be buffered from the tempfile instead
        self.sync()
        return iter(self._decode(iter(self._file.readline, b"")))

    def __enter__(self) -> "AbcStream":
        if "w" in self.mode:
            self._clear_buffer()
            return self

        if "a" in self.mode:
            self.sync()
            self.seek(0, 2)  # go to end of stream
            return self

        self.seek(0)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _write(self, chunk: bytes) -> int:
        ret = self._file.write(chunk)
        if self._pipes:
            self._emit(chunk, self._pipes)
        return ret

    def _clear_buffer(self) -> IO[bytes]:
        if self._buffer and not self._buffer.closed:
            self._buffer.flush()
            self._buffer.close()

        self._buffer = tempfile.SpooledTemporaryFile(
            max_size=self._inmem_size or self.INMEM_SIZE,
            mode="w+b",  # always binary mode
            buffering=self.buffering,
            newline=self.newline,
        )
        return self._buffer

    def _decode(
        self, iter_bytes: Iterable[bytes], encoding: str = None
    ) -> Iterable[str]:
        return (data.decode(encoding or self.encoding) for data in iter_bytes)

    def _teardown(self, signum: signal.Signals, frame):
        self._cleanup()
        if signum == signal.SIGTERM and callable(self._sigterm_handler):
            self._sigterm_handler(signum, frame)
        elif signum == signal.SIGINT and callable(self._sigint_handler):
            self._sigint_handler(signum, frame)
        elif signum == signal.SIGQUIT and callable(self._sigquit_handler):
            self._sigquit_handler(signum, frame)

    def _cleanup(self):
        self._has_read = False
        self._has_stats = False
        if self._file and not self._file.closed:
            self._file.flush()
            self._file.close()

    def __repr__(self):
        type_ = type(self)
        return (
            f"<{type_.__module__}.{type_.__qualname__} "
            f"uri='{self.uri}' mode='{self.mode}' etag='{self.etag}' >"
        )

    def __str__(self):
        with peek_stream(self._file, peek=0) as stream:
            return str(stream.read().decode(self.encoding))

    def __hash__(self):
        identifier = self.uri + self.info.etag
        return hash(identifier.encode())

    def __eq__(self, obj):
        if isinstance(obj, AbcStream):
            return hash(obj) == hash(self)
        return False


AbcStream.register(io.IOBase)
